//
// Monitor for searching for n biggest transients in prefiltered
// data from an arbitrary channel.
//
// Edward Daw, 3rd July 2001

#ifndef __CINT__

#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <string>

#include "DatEnv.hh"
#include "TSeries.hh"
#include "FSeries.hh"
#include "Dacc.hh"
#include "Interval.hh"
#include "Time.hh"
#include "DVector.hh"
#include "TSeries.hh"
#include "DecimateBy2.hh"
#include "IIRFilter.hh"
#include "MonServer.hh"

#include "SlopeFilter.hh"
#include "TransientLeague.hh"

#include "xsil/Xwriter.hh"
#include "xsil/xsil.hh"

#include "TrigBase.hh"
#include "SlopeMonitor.hh"
#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/PeakMon/SlopeMonitor.cc,v 1.5 2003/04/12 01:14:45 jzweizig Exp $"
#define PIDTITLE  "PeakMon"
#include "ProcIdent.hh"

//--------------------------------------  Generate the main routine.
EXECDAT(SlopeMonitor)
#endif               // !def(__CINT__)

using namespace std;

  
//----------------------------- Needed if the monitor exports
//----------------------------- to the DMT viewer
void 
SlopeMonitor::Attention(void)
{
  MonServer::Attention();
}

//--------------------------------------  Skeleton object constructor.
SlopeMonitor::SlopeMonitor(int argc, const char *argv[]) 
  : DatEnv(argc, argv), MaxFrame(999999)
{ 
  // read input arguments

  int sampling_rate=0;

  // cout << "reading input arguments" << endl;
  char eseg_store[100];
  // set default name for monitor server
  // sprintf(monserver_name,"test_peak_monitor");
  monserver_name = "test_peak_monitor";

  // send setup string to logfile
  cout << "monitor started with command: " << endl;
  cout << argv[0] << endl;

  double nyqc = 0;

  if(argc < 17) {
    DumpHelp();
  } else {
    for(int ac = 1; ac < argc; ++ ac) {
      if( strcmp(argv[ac],"-of") == 0 ) {
	sprintf(filepath,"%s",argv[++ac]);
      } else if ( strcmp(argv[ac],"-c") == 0 ) {
	sprintf(channelname,"%s",argv[++ac]);
      } else if ( strcmp(argv[ac],"-mons") == 0 ) {
	monserver_name = argv[++ac];
	//sprintf(monserver_name,"%s",argv[++ac]);
      } else if ( strcmp(argv[ac],"-n") == 0 ) {
	sscanf(argv[++ac],"%d",&no_of_transients);
      } else if ( strcmp(argv[ac],"-f") == 0 ) {
	sprintf(filttype,"%s",argv[++ac]);
      } else if ( strcmp(argv[ac],"-eseg") == 0 ) {
	sprintf(eseg_store,"%s",argv[++ac]);
      } else if ( strcmp(argv[ac],"-sr") == 0 ) {
	sscanf(argv[++ac],"%d",&sampling_rate);
      } else if ( strcmp(argv[ac],"-tldur") == 0 ) {
	sscanf(argv[++ac],"%f",&tldur);
      } else if ( strcmp(argv[ac],"-pkdur") == 0 ) {
	sscanf(argv[++ac],"%f",&pkdur);
      } else if ( strcmp(argv[ac],"-nyqc") == 0 ) {
	sscanf(argv[++ac],"%lf",&nyqc);
      } else if ( isDatEnvArg(argv[ac]) ) {  // data environment args
	++ac; 
      } else {
	DumpHelp();
      }
    }
  }
 
  // set the name of the monitor server
  const char* msn = monserver_name.c_str();
  setServerName(msn);

  cout << "filepath: " 
       << filepath << endl
       << "channel name: " 
       << channelname << endl
       << "number of transients: " 
       << no_of_transients << endl
       << "filter type: " 
       << filttype << endl
       << "sampling rate " << sampling_rate << "Hz" << endl
       << "league duration: " << tldur << "s" << endl
       << "peak duration: " << pkdur << "s" << endl;

  //---------------------------------- allocate memory for array of
  //---------------------------------- pointers to timeseries
  // peakLibrary = new TSeries* [no_of_transients];
  filteredEvent = new TSeries* [no_of_transients];

  //----------------------------------  figure out whether to write event
  //----------------------------------  segment for each event to a file
  if( strcmp(eseg_store,"yes") == 0 ) {
    write_event_segment = true;
  } else {
    write_event_segment = false;
  }

  //----------------------------------  calculate ideal league duration

  int twoexpleaguedur;
  frexp((double)tldur,&twoexpleaguedur);
  transient_league_duration = pow(2,(double)twoexpleaguedur);

  //---------------------------------   calculate ideal segment length
  peakspersegment=64;
  int twoexpseglen;
  frexp((double)pkdur*(double)peakspersegment,&twoexpseglen);
  data_segment_duration = pow(2,(double)twoexpseglen);
 
  //---------------------------------   calculate ideal data rate

  binsperpeak = 8;
  int twoexpbpp;
  frexp((double)pkdur/(double)binsperpeak,&twoexpbpp);
  target_sampling_period = pow(2.0, double(twoexpbpp));

  //---------------------------------   calculate number of decimation stages

  frexp((double)sampling_rate*(double)target_sampling_period,&ndecstages);
  ndecstages -= 1;

  //---------------------------------  if the number of decimation
  //---------------------------------  stages reduces the sampling
  //---------------------------------  rate to below the nyquist cutoff 
  //---------------------------------  for the filter, set the nyquist
  //---------------------------------  filter cutoff manually.
  double nyqu_try = sampling_rate / pow(2.0, ndecstages);
  if( nyqu_try < nyqc ) {
    // reset parameters for manual choice of nyquist cutoff
    target_sampling_period = 1/(2*nyqc);
    frexp((sampling_rate*target_sampling_period),&ndecstages);
    ndecstages -= 1;
  }

  //---------------------------------   calculate number of bins under peak

  peak_nbins = (int)(pkdur/target_sampling_period);

  //---------------------------------   print calculated parameters

  cout << "league time: " << transient_league_duration << "s" << endl 
       << "segment time: " << data_segment_duration << "s" << endl
       << "target samp interval: " << target_sampling_period << "s" << endl
       << "peak width: " << peak_nbins << " bins." << endl
       << "number of decimation stages: " << ndecstages << endl;

  //----------------------------------  set the segment length
  //float tStep(32.0);
  float tStep((float)data_segment_duration);
  getDacc().setStride(tStep);

  //-------------------  set channel name.
  getDacc().addChannel(channelname); 

  //-------------------  set data quality flag.
  last_data_ok = true;

  //-------------------  declare decimator
  if(sampling_rate == 0) {
    cout << "WARNING: setting default number of decimation stages (5)" 
	 << endl;
    DecimateBy2 decimate_to_eight_hz(5,1);
    mydecimate = decimate_to_eight_hz;
  } else {
    //    int ndec = 0;
    //    frexp((double)sampling_rate/8,&ndec);
    //    ndec -= 1;
    //    DecimateBy2 decimate_to_eight_hz(ndec,1);
    DecimateBy2 decimate_filter(ndecstages,1);
    mydecimate = decimate_filter;
  }

  //-------------------  declare filter

  dComplex poles[16];
  dComplex zeros[16];
  target_sampling_rate = 1/target_sampling_period;
  if(strcmp(filttype,"ellip") == 0) {
    GetEllip(poles,zeros,&filter_gain);
    // myiir.init(16,poles,16,zeros,8,filter_gain);
    myiir.init(16,poles,16,zeros,target_sampling_rate,filter_gain);
  } else if(strcmp(filttype,"resgain") == 0) {
    GetResGain(poles,zeros,&filter_gain);
    // myiir.init(2,poles,1,zeros,8,filter_gain);
    myiir.init(2,poles,1,zeros,target_sampling_rate,filter_gain);
  } else if(strcmp(filttype,"Sixtyfive") == 0) {
    GetSixtyfive(poles,zeros,&filter_gain);
    myiir.init(2,poles,1,zeros,target_sampling_rate,filter_gain);
  } else if(strcmp(filttype,"Twohundred") == 0) {
    GetTwohundred(poles,zeros,&filter_gain);
    myiir.init(2,poles,1,zeros,target_sampling_rate,filter_gain);
  } else if(strcmp(filttype,"Sixfifty") == 0) {
    GetSixfifty(poles,zeros,&filter_gain);
    myiir.init(2,poles,1,zeros,target_sampling_rate,filter_gain);
  } else if(strcmp(filttype,"Twothousand") == 0) {
    GetTwothousand(poles,zeros,&filter_gain);
    myiir.init(2,poles,1,zeros,target_sampling_rate,filter_gain);
  } else if(strcmp(filttype,"asqres") == 0) {
    GetAsqResGain(poles,zeros,&filter_gain);
    myiir.init(2,poles,1,zeros,target_sampling_rate,filter_gain);
  } else if(strcmp(filttype,"slope") == 0) {
    // cout << "setting slope filter" << endl;
    slfilt.init(peak_nbins,target_sampling_rate); // order,sampling rate 
    filter_gain = 1.0;
    // cout << "set slope filter" << endl;
  } else if(strcmp(filttype,"nothing") == 0) {
    filter_gain = 1.0;
  } else {
    cout << "ERROR: no such filter" << endl;
    exit(0);
  }

  // cout << "end of monitor initialization" << endl;
  
}


//--------------------------------------  Dump useage screen
void
SlopeMonitor::DumpHelp(void) {
  cout << "useage: PeakMon -of <output file path>               /" << endl
       << "                -c <channel name>                    /" << endl
       << "                -sr <channel sampling rate>          /" << endl
       << "                -eseg <yes / no> to long summary     /" << endl
       << "                -n <number of transients>            /" << endl
       << "                -f <asqres, ellip ,resgain, slope    /" << endl
       << "                   Sixtyfive, Twohundred, Sixfifty   /" << endl
       << "                   Twothousand>                      /" << endl
       << "                -tldur <league duration (s)>         /" << endl
       << "                -pkdur <peak duration (s)>           /" << endl
       << "                -nyqc <manual nyquist cutoff (Hz)>   /" << endl
       << "                -mons <monitor viewer name>          " << endl;

  exit(0);
  return;
}

void
SlopeMonitor::GetResGain(dComplex* poles, dComplex* zeros, double* pfgain) {
  dComplex zero(0.0,0.0);
  dComplex pole0(-0.5,1.732);
  dComplex pole1(-0.5,-1.732);
  poles[0] = pole0;
  poles[1] = pole1;
  zeros[0] = zero;
  *pfgain = 6.2832;
  return;
}

void
SlopeMonitor::GetAsqResGain(dComplex* poles, dComplex* zeros, double* pfgain)
{
  dComplex zero(0.0,0.0);
  dComplex pole0(-35,35.5);
  dComplex pole1(-35,-35.5);
  poles[0] = pole0;
  poles[1] = pole1;
  zeros[0] = zero;
  *pfgain = 6.2832*1.4142*50;
  return;
}

void
SlopeMonitor::GetSixtyfive(dComplex* poles, dComplex* zeros, double* pfgain)
{
  dComplex zero(0.0,0.0);
  dComplex pole0(-45.96,45.96);
  dComplex pole1(-45.96,-45.96);
  poles[0] = pole0;
  poles[1] = pole1;
  zeros[0] = zero;
  *pfgain = 6.2832*1.4142*65;
  return;
}

void
SlopeMonitor::GetTwohundred(dComplex* poles, dComplex* zeros, double* pfgain)
{
  dComplex zero(0.0,0.0);
  dComplex pole0(-141.4,141.4);
  dComplex pole1(-141.4,-141.4);
  poles[0] = pole0;
  poles[1] = pole1;
  zeros[0] = zero;
  *pfgain = 6.2832*1.4142*200;
  return;
}

void
SlopeMonitor::GetSixfifty(dComplex* poles, dComplex* zeros, double* pfgain)
{
  dComplex zero(0.0,0.0);
  dComplex pole0(-459.6,459.6);
  dComplex pole1(-459.6,-459.6);
  poles[0] = pole0;
  poles[1] = pole1;
  zeros[0] = zero;
  *pfgain = 6.2832*1.4142*650;
  return;
}

void
SlopeMonitor::GetTwothousand(dComplex* poles, dComplex* zeros, double* pfgain)
{
  dComplex zero(0.0,0.0);
  dComplex pole0(-1414.2,1414.2);
  dComplex pole1(-1414.2,-1414.2);
  poles[0] = pole0;
  poles[1] = pole1;
  zeros[0] = zero;
  *pfgain = 6.2832*1.4142*2000;
  return;
}

void 
SlopeMonitor::GetEllip(dComplex* poles,dComplex* zeros,double* pfgain) {
  // autogenerated by matlab script makefilt.m. Ed Daw, 9th August 2001
  dComplex zero0(-0.0000000000000,10.7920658694086);
  zeros[0]=zero0;
  dComplex zero1(-0.0000000000000,-10.7920658694086);
  zeros[1]=zero1;
  dComplex zero2(0.0000000000000,4.5973597444516);
  zeros[2]=zero2;
  dComplex zero3(0.0000000000000,-4.5973597444516);
  zeros[3]=zero3;
  dComplex zero4(0.0000000000000,3.6863941562424);
  zeros[4]=zero4;
  dComplex zero5(0.0000000000000,-3.6863941562424);
  zeros[5]=zero5;
  dComplex zero6(-0.0000000000000,3.4464200457381);
  zeros[6]=zero6;
  dComplex zero7(-0.0000000000000,-3.4464200457381);
  zeros[7]=zero7;
  dComplex zero8(0.0000000000000,0.8704684745871);
  zeros[8]=zero8;
  dComplex zero9(0.0000000000000,-0.8704684745871);
  zeros[9]=zero9;
  dComplex zero10(-0.0000000000000,0.8138033733913);
  zeros[10]=zero10;
  dComplex zero11(-0.0000000000000,-0.8138033733913);
  zeros[11]=zero11;
  dComplex zero12(0.0000000000000,0.6525484553652);
  zeros[12]=zero12;
  dComplex zero13(0.0000000000000,-0.6525484553652);
  zeros[13]=zero13;
  dComplex zero14(0.0000000000000,0.2779819949491);
  zeros[14]=zero14;
  dComplex zero15(0.0000000000000,-0.2779819949491);
  zeros[15]=zero15;
  dComplex pole0(-0.0317161385140,2.9974971026841);
  poles[0]=pole0;
  dComplex pole1(-0.0317161385140,-2.9974971026841);
  poles[1]=pole1;
  dComplex pole2(-0.1075281922519,2.8516955944266);
  poles[2]=pole2;
  dComplex pole3(-0.1075281922519,-2.8516955944266);
  poles[3]=pole3;
  dComplex pole4(-0.2002127527594,2.5130217676077);
  poles[4]=pole4;
  dComplex pole5(-0.2002127527594,-2.5130217676077);
  poles[5]=pole5;
  dComplex pole6(-0.2468379143377,1.9889165109294);
  poles[6]=pole6;
  dComplex pole7(-0.2468379143377,-1.9889165109294);
  poles[7]=pole7;
  dComplex pole8(-0.1843579190482,1.4854788823638);
  poles[8]=pole8;
  dComplex pole9(-0.1843579190482,-1.4854788823638);
  poles[9]=pole9;
  dComplex pole10(-0.0945088748225,1.1862524059431);
  poles[10]=pole10;
  dComplex pole11(-0.0945088748225,-1.1862524059431);
  poles[11]=pole11;
  dComplex pole12(-0.0396114033459,1.0505120754334);
  poles[12]=pole12;
  dComplex pole13(-0.0396114033459,-1.0505120754334);
  poles[13]=pole13;
  dComplex pole14(-0.0105885233324,1.0007229599126);
  poles[14]=pole14;
  dComplex pole15(-0.0105885233324,-1.0007229599126);
  poles[15]=pole15;
  *pfgain = 0.0001000431389;
  return;
}

//--------------------------------------  Skeleton object destructor.
SlopeMonitor::~SlopeMonitor() 
{
  delete [] filteredEvent;
  cout << "SlopeMonitor is finished" << endl;
}

//--------------------------------------  Frame processing function.
//-------------------------------------- do this step every timestride.
void
SlopeMonitor::ProcessData(void) {

  // cout << " starting process data " << endl;

  TSeries* pts = getDacc().refData(channelname);

  // check integrity of data for decimator

  try {
    mydecimate.dataCheck(*pts);
  } catch (exception& de) {
    last_data_ok = false;
    mydecimate.reset();
    return;
  }
  
  // decimate data
  // cout << "decimating" << endl;
  TSeries decimated(mydecimate.apply(*pts));

  // filter the data

  // cout << "filtering" << endl;
  TSeries filtered;
  if( (strcmp(filttype,"ellip") == 0) ||
      (strcmp(filttype,"resgain") == 0) ||
      (strcmp(filttype,"Sixtyfive") == 0) ||      
      (strcmp(filttype,"Twohundred") == 0) ||
      (strcmp(filttype,"Sixfifty") == 0) ||
      (strcmp(filttype,"Twothousand") == 0) ) {
    filtered = myiir.apply(decimated);
  } else if (strcmp(filttype,"nothing") == 0) {
    filtered = decimated;
  } else {
    filtered = slfilt.apply(decimated);
  }

  // multiply filtered data by gain unless using the resgain filter
  // check that I nevertheless need to multiply by gain for the non
  // resgain filters
  if ( (strcmp(filttype,"ellip") == 0) ||
       (strcmp(filttype,"slope") == 0) ) {
    filtered *= filter_gain;
  }
  
  // search routine
  TSeries raw(*pts);

  if( !transient_store.inUse() ) {

    transient_store.initialize(filtered,raw,no_of_transients,peak_nbins,
			       2*peak_nbins, monserver_name);
    last_data_ok = true;
    // cout << "Initialized transient store" << endl;

  } else {

  // search for peaks in the new data

    bool is_data_contiguous = true;
    cout << "searching" << endl;
    if(last_data_ok) {
      try {
	cout << "searching full" << endl;
	transient_store.search(filtered,raw,0,is_data_contiguous);
      } catch (exception& e) {
	last_data_ok = false;
      }
    }
    else {
      last_data_ok = true;
      is_data_contiguous = false;
      try {
	cout << "searching missed" << endl;
	int searcherror = 
	  transient_store.search(filtered,raw,transient_store.getHistoryLength(),
				 is_data_contiguous);
	if( searcherror == -1 ) {
	  last_data_ok = false;
	}
     } catch (exception& e) {
	last_data_ok = false;
      }
    }
    
    // write transiets to data monitor viewer
    char evName[50];
    char rawName[50];
    char all_cumulative_filtered_hist_name[20]={"cumulative_hist"};
    char all_hourly_filtered_hist_name[20]={"period_hist"};
    serveData(all_cumulative_filtered_hist_name,
	      transient_store.getHistPointer( ALL_CUMULATIVE_FILTERED ));
    serveData(all_hourly_filtered_hist_name,
	      transient_store.getHistPointer( ALL_HOURLY_FILTERED ));
    for( int i=0 ; i < no_of_transients ; ++i ) {
      sprintf(evName,"filtered_event_%d",i);
      sprintf(rawName,"raw_event_%d",i);
#ifdef DEBUGSLOPE
      cout << "serving data" << endl;
#endif
      serveData(evName,transient_store.monitorUpdate(i));
      serveData(rawName,transient_store.monitorRawUpdate(i));
    }

    // write the transient summary tables to a file
    //
    sprintf(full_filename,"%s/%lu.trans",
	    filepath,
	    (transient_store.startTime()).getS() );
    ofstream outfile;
    outfile.exceptions( ofstream::eofbit | 
			ofstream::failbit | 
			ofstream::badbit );
    // try and open file
    try {
      cout << "opening and writing file" << endl;
      outfile.open(full_filename);
      transient_store.outputLeague(outfile,write_event_segment);
      outfile.close();
    } catch(exception& e) {
      cerr << "cannot open file for output" << endl;
    } 
   
    // print start time info.
    cout << "league start time " << transient_store.startTime() << endl;
    cout << "current start time " << pts->getStartTime() << endl;
    cout << "league duration " << Interval(transient_league_duration) << endl;

    // if more than league_time_secs seconds have elapsed, reset the
    // transient league
    
    if( ( pts->getStartTime() - transient_store.startTime() )
	>= Interval(transient_league_duration) ) {
      
      // write the last hours list of triggers.
      //int ntriggers = transient_store.getNEvents();
      //cout << "getting " << ntriggers << " triggers" << endl;
      //for(int i=0; i<ntriggers; ++i) {
      // 	trig::TrigBase trigData("PeakMon",channelname);
      // 	trigData.setTime(transient_store.getEventTime(i));
      // 	trigData.setDuration(transient_store.getEventDuration(i));
      // 	trigData.setPriority(trig::p_info);
      // 	trigData.setDisposition(trig::d_metaDB);
      // 	int rc = sendTrigger(trigData);
      // 	if(!rc) {
      // 	  cout << "error sending trigger" << endl;
      // 	}
      // }

      // reset transient store
      cout << "Emptying event list" << endl;
      transient_store.RestartEventList(pts->getEndTime());
      cout << "Event list emptied" << endl;

    }
  }

  return;
}


