#include "absGlitch.hh"
#include "ParseLine.hh"

#include <iostream>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <iomanip>
#include <string>
#include <vector>

using namespace std;

EXECDAT(absGlitch);
   
  
// constructor
absGlitch::absGlitch(int argc, const char *argv[]) : DatEnv(argc,argv), 
  MonServer("absGlitch"), mOSC(getDacc()), fData(0), fFilterData(0),
  fCurrentTime(0), fLogStartTime(0), fLogInit(false), fDataInit(false)
{
  if ( Debug() > 1 ) cout << "Starting absGlitch..."<<endl;
  const char *confpath = getenv("DMTPARS");
  string cfile, oscfile;
  if (confpath) {
    cfile   = (string)confpath + "/absGlitch.conf";
    oscfile = (string)confpath + "/LockLoss.conf";
  } else {
    cfile   = "./absGlitch.conf";
    oscfile = "./absGlitch_osc.conf";
  }

  Reset();
  
  //=======================  Parse command line arguments
  for (int i=1; i<argc; i++) {
    if (!strcmp("-conf", argv[i])) {
      if (chkNextArg(i,argv,argc)) cfile = argv[++i];
      else {
	cerr << "Couldn't find configuration file"<<endl;
	exit(Usage(argv[0])+1);
      }
    } else if (!strcmp("-osc", argv[i])) {
      if (chkNextArg(i,argv,argc)) oscfile = argv[++i];
      else {
	cerr << "Couldn't find OSC file"<<endl;
	exit(Usage(argv[0])+1);
      }
    } else if (!strcmp("-h",argv[i])|| !strcmp("--help",argv[i])) {
      exit(Usage(argv[0]) - Help());
    } else if (isDatEnvArg(argv[i])) i++;
    else exit(Usage(argv[0]));
  }

  //=======================  Define and initialize parameters
  mDict.addPar("Data_File_Prefix",  std::string("absGlitch_data"));
  mDict.addPar("Log_File_Prefix",   std::string("absGlitch_log"));
  mDict.addPar("Oper_State_Cond",   std::string(""));
  mDict.addPar("Number_Bins",       std::string("20"));
  mDict.addPar("Total_Runtime",     std::string("0"));
  mDict.addPar("Log_Interval",      std::string("86400"));
  mDict.addPar("Trigger_ON/OFF",    std::string("1"));
  mDict.addPar("Trigger_Bin",       std::string("1"));
  
  //=======================  open & read config file
  SetConfig(cfile.c_str());
  if (mDict.getString("Oper_State_Cond").size() > 2) mOSC.readConfig(oscfile);
  getDacc().setIgnoreMissingChannel(true);

  //=======================  start the log file
  OutputLog(true);

  //=======================  allocate buffer
  fData        = new float[16384 * 2];
  fFilterData  = new float[16384 * kOffsetFactor];
}


// destructor
absGlitch::~absGlitch(void) 
{
  if ( Debug() > 1) cout << "Closing absGlitch..." << endl;
  OutputLog(false);
  Time lastT(fCHStatus.begin()->fTSnew->getEndTime());

  //------- Flush all events
  for (CSiter ch = fCHStatus.begin(); ch != fCHStatus.end(); ch++) {
    if (ch->HandleEvent(0.0, lastT, true) ) OutputEvent(ch);
  }

  //========= Close files
  if (fLogFile.is_open()) {
    fLogFile << "Monitor End = " <<lastT.getS() <<endl
	     << "====================================" <<endl;
    fLogFile.close();
  }
  if (fDataFile.is_open()) fDataFile.close();

  //========= Free memory
  if (fData) delete[] fData;
  if (fFilterData) delete[] fFilterData;
}


void absGlitch::SetConfig(const char *const config) 
{
  if (Debug() > 1) cout <<"Reached ReadConfig(void)"<<endl;
  //=======================  Set up the line parser
  ParseLine lp(config);
  if (Debug()) lp.setLog(cout);
  int nLine;
     
  //=======================  read and set configurations
  for (int nArg=lp.getLine(); nArg>=0; nArg=lp.getLine()) {
    nLine++;
  //-----------  Empty command line
    if (!lp[0]) {
      continue;
  //-----------  Parse parameters
    } else if (nArg < 4) {
      if (!mDict.exists(lp[0])) {
	cerr<<"No such parameter: "<<lp[0]<<endl;
	break;
      } else {
	if (Debug()) cout<<"Set parameter: "<<lp[0]<<" = "<<lp[nArg - 1]<<endl;
	mDict.setPar(lp[0], lp[nArg - 1]);
      }
  //-----------  Parse channels
    } else if ( !strcmp(lp[0], "Channel") ) {
      ChannelStatus chstat;
      chstat.setDebug(Debug());
      chstat.fName           = lp[1];
      //      chstat.fSampleFreq     = atoi(lp[2]);
      chstat.fDeadTime       = atof(lp[2]);
      chstat.fUseOSC         = atoi(lp[3]);
      chstat.fRMS            = atof(lp[4]);
      chstat.fThreshold      = atof(lp[5]);
      switch (nArg) {
      case 7:
	chstat.fTrigComment = lp[6];
	break;
      case 8:		// We'll also set trigger comment == channel name
	chstat.SetupCoefs( lp[6], lp[7], "", "Hz", Debug() );
      case 6:
	chstat.fTrigComment = lp[1];
	break;
      case 9:
      case 10:
      case 11:
	chstat.fTrigComment = lp[1];
        string units = "Hz";
	string gain  = "1.0";
        for (int arg = 8; arg < nArg; arg++) {
	  if ( !strcmp(lp[arg],"Hz") || !strcmp(lp[arg],"rad/s") ) {
	    units = lp[arg];
	  } else if ( strpbrk(lp[arg],":_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz") ) {
	    chstat.fTrigComment = lp[arg];
	  } else {
	    gain = lp[arg];
	  }
	}
	chstat.SetupCoefs( lp[6], lp[7], gain.c_str(), units.c_str(), Debug());
	break;
      }
      fCHStatus.push_back(chstat);
      getDacc().addChannel( chstat.fName.c_str() );
    }
  }

   
  //=======================  Set up member elements
  fBinNumber        = atoi(mDict.getString("Number_Bins").c_str());
  fTriggerON        = atoi(mDict.getString("Trigger_ON/OFF").c_str());
  fTriggerBin       = atoi(mDict.getString("Trigger_Bin").c_str());
  if (fTriggerBin > fBinNumber) {
    cerr <<"Bin to trigger on is greater than the total number of bins"
	 <<"-- setting to max"<<endl;
    fTriggerBin = fBinNumber;
  }
  fEndRuntime       = atoi(mDict.getString("Total_Runtime").c_str());
  fLogInterval      = atof(mDict.getString("Log_Interval").c_str());
  fOSCName          = mDict.getString("Oper_State_Cond");
  // reset OSCname if uncommented by accident
  if (fOSCName.size() < 3 ) fOSCName = "";
  fLogPath          = mDict.getString("Log_File_Prefix");
  fDataPath         = mDict.getString("Data_File_Prefix");

  // debug whether there's anything to analyze or not
  fNChannel = fCHStatus.size();
  if (!fNChannel) { // kill if fNChannel or fFilterOrder <=0
    if (!fNChannel) cerr << "# of Channels <= 0" << endl;
    finish();
    return;
  }

  getDacc().synch();
  getDacc().fillData(Interval(1.0));
  // allocate time series and histogram and set channels to data access,
  // then grab a sample time series to fill in the rest of the info.
  for (CSiter i = fCHStatus.begin(); i != fCHStatus.end(); ++i) {
    if ( getDacc().isChannelRead(i->fName.c_str()) ) {
      i->fSampleFreq = getDacc().refData(i->fName.c_str())->getNSample();
      
      // Need to construct 'float' TSeries (3 args defaults to float type)
      i->ResetFilter( i->fPoles, i->fZeroes, i->fSampleFreq, i->fGain );

      if ( Debug() > 1 ) cout << "Some Channel info: " << i->fName<<" "
			      << i->fSampleFreq<<" "
			      << i->fUseOSC<<" "
			      << i->fThreshold<<" "
			      << i->fDeadTime<<endl;
      // If threshold is rather low compared to RMS, we should modify the bin
      // size.  Otherwise we'll just let the RMS determine the bin size.
      if ( (i->fThreshold - i->fRMS) >= i->fRMS )
	i->ResetHistograms( fBinNumber, i->fThreshold, 
			    i->fThreshold + fBinNumber * i->fRMS );
      else 
	i->ResetHistograms( fBinNumber, i->fThreshold, 
			    i->fThreshold + fBinNumber * 
			                    (i->fThreshold - i->fRMS) );
      string name = i->fName + "_absGlitch";
      string nameall = i->fName + "_ALL_absGlitch";	
      MonServer::serveData(name.c_str(),&(i->fHist));
      MonServer::serveData(nameall.c_str(),&(i->fHistAll));
    } // if ( getDacc().isChannelRead(i->fName.c_str()) )
    else cerr << "Channel " << i->fName << " not found." << endl;
  } // for (CSiter i = fCHStatus.begin(); i != fCHStatus.end(); ++i)

  // debug whether we've hooked up with data stream or not
  fCurrentTime = fInitTime = getDacc().getCurrentTime();
  if ( fCurrentTime == Time(0) ) cerr << "Couldn't find current time." <<endl;
}


// main loop
void absGlitch::ProcessData(void)
{
  fCurrentTime = getDacc().getCurrentTime();
  if ( Debug() > 1 ) cout <<"Processing Data: "<<fCurrentTime.getS()<<endl;
  //=======================  If data is continuous, update stats & check event
  if ( fCHStatus.begin()->IsTSContinuous() ) { // check continuity
    for (CSiter ch = fCHStatus.begin(); ch != fCHStatus.end(); ch++) {
      //----------- If the OSC, if used, is satisfied...
      if ( ch->fUseOSC  && (Debug() > 1) )
	 cout << "Using osc: " << fOSCName << "-- value: "
	      << mOSC.watchedQuantity(fOSCName) << "; satisfied? "
	      << mOSC.satisfied(fOSCName) << endl;
      if ( !(ch->fUseOSC) || mOSC.satisfied(fOSCName.c_str()) ) {
	//--------- convert TSeries to double (note: should take 0 time, but
	//---------  we'll be safe (if slow) by doing it here.  Note that
	//---------  the reason we need this is IIRFilter only uses doubles
	getDacc().refData( ch->fName.c_str() )->Convert(DVector::t_double);
	//--------- filter data
	ch->FilterData( *(getDacc().refData( ch->fName.c_str() )) );
	//--------- update the filtered data statistics
	ch->UpdateStats();
	//--------- check for event
	CheckEvent( ch );
      } // if ( (fOSCName != "") || !(ch->fUseOSC) || ...
    } // for (CSiter ch = fCHStatus.begin(); ...

  //=======================  Otherwise there's a continuity error
  } else {
    fLogFile<<"TimeSeries Continuity Error "
	    <<fCHStatus.begin()->fTSnew->getStartTime().getS()
	    <<endl;
  } // if (fCHStatus.begin()->IsTSContinuous())

  //=======================  Output stats, if need be
  OutputLog(true);
  
  //=======================  Check if we've reached the end of the run
  fCount++; // increment # of ProcessData executed
  if (fEndRuntime) {
    if (fCount >= fEndRuntime) finish();
  }
}

// data holds filtered time series
void absGlitch::CheckEvent( ChannelStatus *ch ) {
  // reject channel with rms less than 1 count (saturated)
  if (ch->fRMS > 1) {
    for (int j=0; j < ch->fSampleFreq; j++) {
      Time bintime = ch->fTSnew->getBinT(j);
      const double datapt = ch->fTSnew->getDouble(j);
      
      //------- Check whether datapt ends an event
      if (ch->HandleEvent(datapt,bintime) ) OutputEvent(ch);
    }
  }
}

//output glitchdata data
void absGlitch::OutputEvent(ChannelStatus* ch) 
{
  //==========================  Repeat for all thresholds
  for (unsigned int i=1; i < ch->fEventStatus.size(); i++) {
    ChannelStatus::absEvent *event = &(ch->fEventStatus[i]);
    //===================  First see that event has finished
    if ( event->status == ChannelStatus::kDone ) {
      if (Debug() > 1)
	cout <<"Outputting Event... channel " <<ch->fName 
	     <<"\tthreshold= " <<ch->fHist.GetBinLowEdge(i)
	     <<"\tfEventMax= " <<event->amplitude <<endl;
      //:NOTE: include the next couple lines if you wish to put dead space
      //:NOTE: between the actual event and when the data is good.
      // event->startTime -= (Interval)ch->fDeadTime;
      // event->endTime   += (Interval)ch->fDeadTime;

      //=====================  Setup trigger values
      double duration    = event->endTime - event->startTime;
      double snglBinSize = ( ch->fPoles.size() + ch->fZeroes.size() +.5 ) /
                           (2 * ch->fSampleFreq) ;
      //=====================  Add event to histograms
      if ( !(ch->fDeadTime) || (duration > snglBinSize) ) {
	ch->fHist.Fill( ch->fHist.GetBinLowEdge(i) );
	ch->fHistAll.Fill( ch->fHist.GetBinLowEdge(i) );
	ch->fTimeHist.Fill( ch->fHist.GetBinLowEdge(i), duration );
	ch->fTimeHistAll.Fill( ch->fHist.GetBinLowEdge(i), duration );
      } else ch->fShortHist.Fill( ch->fHist.GetBinLowEdge(i) );

      //=====================  Add event to data file
      if ( (fTriggerON & 1) && ch->fDeadTime && i == fTriggerBin ) {
	//-----------  Initialize data file if it's not already
	if ( !fDataInit ) {
	  // append initial UTC date to the filename & prepend appropriate path
	  string fDataName = fDataPath;
	  char *outSuffix= new char[12];
	  fDataName += (string) TimeStr(fInitTime,outSuffix,".%y.%m.%d");
	  delete[] outSuffix;
	  if (fDataName[0] != '/' && fDataName[0] != '.') {
	    const char* datadir = getenv("DMTOUTPUT");
	    if (datadir) fDataName = string(datadir) + "/" + fDataName;
	  }
	  fDataFile.open(fDataName.c_str(),ios::out|ios::app);
	  if (!fDataFile.is_open()){
	    cerr<<"Data File Error: "<<fDataName<<endl;
	    finish();
	    return;
	  } 
	  fDataFile << "### Trigger Statistics:" << endl
		    << "### Event Start | Channel | Event Start | Duration  | "
		    << "Max Glitch Size | RMS during | Trigger"<<endl
		    << "###    (UTC)    |   Name  |    (GPS)    | (seconds) | "
		    << "  (in counts)   |   Glitch   | Comment"<<endl;
	  fDataInit = true;
	} //  if ( !(fDataInit) )
  
	//----------- output event to data file
	char str[40] = "";
	TimeStr(event->startTime, str, "%y.%m.%d %H:%N:%S ");
	fDataFile <<str << ch->fName<<" "
		  <<setprecision(16) <<double(event->startTime.totalS())<<" "
		  <<setprecision(15) <<duration <<" "
		  <<setprecision(6)  <<event->amplitude <<" "
		  <<setprecision(6)  <<event->variance  <<" "
		  <<ch->fTrigComment <<endl;
      } //  if ( fTriggerON & 1 )
  
      //=====================  Send event to Trigger Manager
      if ( (fTriggerON & 2) && ch->fDeadTime  && i == fTriggerBin ) {
	string subID = ch->fName;
	trig::TrigRslt trigData("absGlitch",subID.c_str());
	trigData.setTime(        event->startTime);
	trigData.setDuration(    duration );
	trigData.setIntensity(   event->amplitude);
	trigData.setSignificance(event->variance);
	trigData.setPriority(    trig::p_warn);
	trigData.setDisposition( trig::d_metaDB);
	if ( ch->fTrigComment.size() )
	  trigData.setData( ch->fTrigComment.c_str(),
			    ch->fTrigComment.size()  );
	int trigerr = sendTrigger(trigData);
	if ( trigerr && fLogFile.is_open() )
	  fLogFile<<"Error Sending Trigger: " <<trigerr <<" at "
		  <<event->endTime <<endl;
      } //end if ( fTriggerON & 2 ) ...
      event->status = ChannelStatus::kNone;

    } //end if ( event->status == ChannelStatus::kDone ) ...
  } //end for (int i=1; i < ch->fEventStatus.size(); i++) ...
  return;
}


// output log file 
void absGlitch::OutputLog(bool check) 
{
  //=======================  Initialize log file if none exists
  if (!fLogInit) {
    // append initial UTC date to the filename & prepend appropriate path
    string fLogName = fLogPath;
    char *outSuffix= new char[12];
    fLogName += (string) TimeStr(fInitTime,outSuffix,".%y.%m.%d");
    delete[] outSuffix;
    if (fLogName[0] != '/' && fLogName[0] != '.') {
      const char* htmldir = getenv("DMTHTMLOUT");
      if (htmldir) fLogName = string(htmldir) + "/" + fLogName;
    }
    fLogFile.open(fLogName.c_str(),ios::out|ios::app);
    if (!fLogFile.is_open()){
      cerr<<"Log File Error: "<<fLogName<<endl;
      finish();
      return;
    }
   
  //-----------  Output configuration
    fLogFile << "#====================================" << endl
	     << "# file path to the text output data file" << endl
	     << "Data_File_Prefix : " << fDataPath << endl << endl
	     << "# length of data to process in seconds (0 for infinity)"<<endl
	     << "Total_Runtime    : " << fEndRuntime << endl << endl
	     << "# of bins which composes each histogram" << endl
	     << "Number_Bins      : " << fBinNumber << endl << endl
	     << "# dispose of triggers (0:none; 1,3:To file; 2,3:To Trigger Manager)"<<endl
	     << "Trigger_ON/OFF   : " << fTriggerON << endl << endl
	     << "# bin to trigger on" <<endl
	     << "Trigger_Bin      : " << fTriggerBin <<endl <<endl
	     << "# file path to the log file" << endl
	     << "Log_File_Prefix  : " << fLogPath << endl << endl
	     << "# time interval to update log in seconds (0 for single update at end)" << endl
	     << "Log_Interval     : " << fLogInterval << endl << endl
	     << "# Operating State Condition to obey" << endl
	     << "Oper_State_Cond  : " << fOSCName << endl << endl
	     << "# of Channels    : " << fCHStatus.size() << endl << endl
	     << "### Configuration (by channel)" << endl
	     << "### Channel | Wait | Use  |   Bin   | Min Thresh  | "
	     << "Filter | Filter |[Filter]|[Freq ]| Trigger "<<endl
	     << "###  Name   | Time | OSC? | Spacing | (in counts) | "
	     << "Poles  | Zeroes |[ Gain ]|[Units]| Comment "<<endl;
    for ( CSiter i = fCHStatus.begin(); i != fCHStatus.end(); ++i ) {
      fLogFile<<"Channel "<<i->fName<<' '	<<i->fDeadTime<<' '
	      <<i->fUseOSC<<' '			<<i->fRMS<<' '
	      <<i->fThreshold<<' ';
      i->DumpPoles(fLogFile);
      i->DumpZeroes(fLogFile);
      fLogFile<<i->fGain<<' '			<<"Hz "
	      <<i->fTrigComment<<endl;
    }

  //-----------  output starting time
    if ( fLogStartTime == Time(0) ) { // output absGlitch start time
      fLogStartTime = fCurrentTime;
      fLogFile << "###===================================="      << endl
	       << "### Monitor Start = " << fLogStartTime.getS() << endl
	       << "###------------------------------------"      << endl;
  //-----------  set histogram time stamps
      for ( CSiter ch = fCHStatus.begin(); ch != fCHStatus.end(); ch++ ) {
	ch->fHist.SetTime(fCurrentTime);
	ch->fHistAll.SetTime(fCurrentTime);
	ch->fTimeHist.SetTime(fCurrentTime);
	ch->fTimeHistAll.SetTime(fCurrentTime);
	ch->fShortHist.SetTime(fCurrentTime);
      }
    } // if ( !fLogStartTime.getS() )
    fLogCount = 0;
    fLogInit = true;
  } // if (!fLogInit)
  
  //=====================  update log file every fLogInterval or at end of run
  if ( !check || 
       (fLogInterval && (fCurrentTime > fLogStartTime + fLogInterval)) ) {
    fLogFile << "### Log_Interval: " << fLogCount
	     << "\tGPS_start: " << fLogStartTime.getS()
	     << "\tInterval_Duration: " << double(fCurrentTime - fLogStartTime)
	     <<endl
	     << "### Channel Statistics" << endl
	     << "### Channel | Filtered | Threshold | # Triggers  | # Triggers"
	     << " | Current  |  Dead Time  | Dead Time | Filtered | # Single-"
	     << endl
	     << "###  Name   |   RMS    |           | in Interval |    Total  "
	     << " | Run Time | in Interval |   Total   | Kurtosis | Bin "
	     << "Trigs" << endl;
    // output channel statistics
    Histogram1::stat_t mHistStats[4];
    for ( CSiter ch = fCHStatus.begin(); ch != fCHStatus.end(); ch++) {
      int nBins = ch->fHist.GetNBins() + 1;
      //=======  Print the total events in each histogram
      fLogFile <<"#"<<fLogCount<<" "                    <<ch->fName<<"\t"
	       <<setprecision(6)<<ch->fRMS<<"\t" <<"Totals"<<"\t";
      ch->fHist.GetStats(mHistStats); // 0th element is sum of events
      fLogFile <<mHistStats[0] + ch->fHist.GetBinContent(nBins)<<"\t";
      ch->fHistAll.GetStats(mHistStats); // 0th element is sum of events
      fLogFile <<mHistStats[0] + ch->fHistAll.GetBinContent(nBins)<<"\t"
	       <<fCount<<"\t";
      ch->fTimeHist.GetStats(mHistStats); // 0th element is sum of events
      fLogFile <<mHistStats[0] + ch->fTimeHist.GetBinContent(nBins)<<"\t";
      ch->fTimeHistAll.GetStats(mHistStats); // 0th element is sum of events
      fLogFile <<mHistStats[0] + ch->fTimeHistAll.GetBinContent(nBins)<<"\t"
	       <<ch->fKurt<<"\t";
      ch->fShortHist.GetStats(mHistStats); // 0th element is sum of events
      fLogFile <<mHistStats[0] + ch->fShortHist.GetBinContent(nBins)<<"\t"
	       <<endl;
      
      //=======  Print the event information for each threshold level
      for (int i=fBinNumber + 1; i > 0 ; i--) //N+1 bins range from [1...(N+1)]
	fLogFile << "#"<<fLogCount<<" "
		 << ch->fName<<"\t"             
		 << setprecision(6)<<ch->fRMS<<"\t"
		 << setprecision(6)<<ch->fHist.GetBinLowEdge(i)<<"\t"
		 << ch->fHist[i]<<"\t"
		 << ch->fHistAll[i]<<"\t"
		 << fCount<<"\t"
      		 << ch->fTimeHist[i]<<"\t"
      		 << ch->fTimeHistAll[i]<<"\t"
		 << setprecision(6)<<ch->fKurt<<"\t"
		 << ch->fShortHist[i]<< endl;
      ch->fStatCount = 0; // This resets the statistics
      // clear contents and update histogram time stamp (fHist only)
      ch->ClearHistogram();
      ch->fHist.SetTime(fCHStatus.begin()->fTSnew->getEndTime());
      ch->fTimeHist.SetTime(fCHStatus.begin()->fTSnew->getEndTime());
      ch->fShortHist.SetTime(fCHStatus.begin()->fTSnew->getEndTime());
    }
    fLogFile <<"###====================================" << endl;
    fLogStartTime = fCurrentTime;
    fLogCount++;
  } // if ( !check || fCurrentTime > fLogStartTime + fLogInterval )
  return;  
}

void absGlitch::Reset()
{
  if (fFilterData) delete fFilterData;
  fFilterData   = 0;
  if (fData) delete fData;
  fData         = 0;
  fMaxData      = 0;
  fNChannel     = 0;
  fEndRuntime   = 0;
  fCount        = 0;
  fLogStartTime = Time(0);
  fLogInterval  = 0;
  fCurrentTime  = Time(0);
  fLogPath      = "";
  fDataPath     = "";
  fOSCName      = "";
  fTriggerON    = 0;
}

// Checks that next argument isn't an option
bool absGlitch::chkNextArg(int i, const char* argv[], const int argc) 
{
  if (++i < argc) {
    std::string args(argv[i]);
    if (args[0] != '-') return true;
  } // Note: Doesn't check next arg for number value
  return false;
}

// Prints out a usage statement
int absGlitch::Usage(const char* progName) 
{
  cerr << "Usage: "<< progName<<" [-conf <file>] [-osc <file>] "
       << "[-infile <frame_file>] [-h | --help]" 
       << endl;
  return 1;
}

// Prints out a Help statement (not helpful)
int absGlitch::Help(void) 
{
  cerr <<"absGlitch looks for glitches with absolute thresholds (in counts)"
       <<"\nOptions include:\n"
       <<"  -conf   <file>       -- The configuration file to use as input\n"
       <<"  -osc    <file>       -- An OSC configuration file\n"
       <<"  -infile <frame_file> -- Used when running offline.  Frame files\n"
       <<"      can include a wildcard character, which must be escaped, as\n"
       <<"      in:   '-infile <frame_path>/H-R-\\*.gwf'\n"
       <<"  -h or --help         -- Generates this help text and exits"<<endl;
  return 1;
}

void absGlitch::Attention() 
{
  MonServer::Attention();
}



