#include <tid.h>

void ChnInfo::ScaleThreshold(float sf) {
  
  unsigned int i;
  for(i=0;i<powThresh.size();i++)
    powThresh[i] *= sf;
  
  eventsList->SetPThresh(powThresh);
}


ChnInfo::ChnInfo (const int cid_,
		  const string& name_,
		  const unsigned int fs_,
		  const double fLow_,
		  const double fHigh_,
		  const unsigned int dec_,
		  const Interval timeRes_,
		  const Interval stabScale_,
		  const double stabTol_,
		  const vector<unsigned int>& testFreq_,
		  const double p_,
		  const unsigned int sigma_,
		  const vector<unsigned int>& delta_,
		  const double alpha_) {    // Constructor

  doFilter = 0;
  OSC = 0;
  cid = cid_;
  name = name_;
  fs = fs_;
  fLow = fLow_;
  fHigh = fHigh_;
  dec = dec_;
  timeRes = timeRes_;
  stabScale = stabScale_;
  stabTol = stabTol_;
  testFreq = testFreq_;
  ForceComputeThreshold = 0;
  mr = mi = s2r = s2i = NULL;

  for(unsigned int i=0; i<testFreq.size(); i++) {
    ListComplex* tlist = new ListComplex;
    History.push_back ( tlist );
  }

  MeanValues = new Complex[fs];
  VarValues = new Complex[fs];

  powThresh.resize(fs,0);

  thrCompMode = 1; // need to acquire data

  bpp = p_;
  sigma = sigma_;
  delta = delta_;
  alpha = alpha_;

  eventsList = new EventsList(name,
			      sigma,
			      delta,
			      alpha,
			      powThresh, // empty vector
			      Diagnostic,
			      fLow,
			      fHigh,
			      timeRes);

  eventsList->SetMaxClusterDuration((unsigned int)floor(stabScale / timeRes));

  Nsum = 0;
  mr = mi = s2r = s2i = NULL;
}


ChnInfo::ChnInfo (const int cid_,
		  const string& name_,
		  const unsigned int fs_,
		  const double fLow_,
		  const double fHigh_,
		  const unsigned int dec_,
		  const Interval timeRes_,
		  const Interval stabScale_,
		  const double stabTol_,
		  const vector<unsigned int>& testFreq_,
		  const double p_,
		  const unsigned int sigma_,
		  const vector<unsigned int>& delta_,
		  const double alpha_,
		  const vector<float>& powThresh_) {    // Constructor
  doFilter = 0;
  OSC = 0;
  cid = cid_;
  name = name_;
  fs = fs_;
  fLow = fLow_;
  fHigh = fHigh_;
  dec = dec_;
  timeRes = timeRes_;
  stabScale = stabScale_;
  stabTol = stabTol_;
  testFreq = testFreq_;
  ForceComputeThreshold = 0;
  mr = mi = s2r = s2i = NULL;

  
  for(unsigned int i=0; i<testFreq.size(); i++) {
    ListComplex* tlist = new ListComplex;
    History.push_back ( tlist );
  }
  
  MeanValues = new Complex[fs];
  VarValues = new Complex[fs];

  powThresh.resize(fs,0);

  thrCompMode = 0; // no need to acquire data for thresholds
  powThresh = powThresh_;

  bpp = p_;
  sigma = sigma_;
  delta = delta_;
  alpha = alpha_;

  eventsList = new EventsList(name,
			      sigma,
			      delta,
			      alpha,
			      powThresh,
			      Diagnostic,
			      fLow,
			      fHigh,
			      timeRes);

  eventsList->SetMaxClusterDuration((unsigned int)floor(stabScale / timeRes));

  Nsum = 0;
  mr = mi = s2r = s2i = NULL;
}



ChnInfo::~ChnInfo () // destructor
{
  testFreq.clear();

  for(unsigned int i=0; i<History.size(); i++) 
    History[i]->clear();
  History.clear();

  delete MeanValues;
  delete VarValues;

  if(mr) delete mr;
  if(mi) delete mi;
  if(s2r) delete s2r;
  if(s2i) delete s2i;

  powThresh.clear();
  delta.clear();

  delete eventsList;

  delete filter;
}


void ChnInfo::ConstructFilter () {

  dComplex *polesptr = new dComplex[poles.size()];
  dComplex *zeroesptr = new dComplex[zeroes.size()];

  for(unsigned int i=0; i<poles.size(); i++) {
    polesptr[i] = poles[i];
    //    cout<<polesptr[i]<<endl;
  }
  for(unsigned int i=0; i<zeroes.size(); i++) {
    zeroesptr[i] = zeroes[i];
    //    cout<<zeroesptr[i]<<endl;
  }

  filter = new IIRFilter(poles.size(), polesptr, zeroes.size(), zeroesptr, (double)fs);

  delete polesptr;
  delete zeroesptr;

}

void ChnInfo::ProcessData(unsigned int cid_,
			  const FSeries* s) { // process one spectrum; cid_ should match cid

  if(cid_ != cid)
    throw TidError ( "FSeries object piped to the wrong container." );

  // set time Id
  eventsList->SetNowId(s->getStartTime());

  // get data
  Complex data[fs/2];
  if(fs/2 != s->getData(fs,data))
    throw TidError ( "FSeries length doesn't match sampling frequency." );


  // store Fourier amplitudes for testpoints
  unsigned int nTSteps = (unsigned int)floor(stabScale / timeRes);

  if(!thrCompMode) {
    for(unsigned int i=0; i<testFreq.size(); i++) {
      History[i]->push_back(data[testFreq[i]]);

      if(History[i]->size() > nTSteps)
	History[i]->pop_front(); // delete oldest entry if list too big
    }
  }

  if(!TestStationarity() || ForceComputeThreshold) {

    if(!ForceComputeThreshold) 
      TidWarning ( string("Channel ") + name + string(" is not stationary over the required timescale.") );
    else {
      TidWarning( string("Threshold for channel ") + name + string(" was requested to be recomputed.") );
      ForceComputeThreshold = 0;
    }

    TidWarning ( string("Recomputing thresholds for channel ") + name );

    ComputeThresholds(s); // thresholds recomputation

    thrCompMode = 1;

  }
  else {
    if(thrCompMode) ComputeThresholds(s); // thresholds recomputation
    else {

      FSpectrum pspec(*s);

      eventsList->Add(cid_, &pspec);

      eventsList->Clean();

      eventsList->Purge();

    }
  }
  
}


void ChnInfo::ComputeThresholds(const FSeries* s) { // compute the thresholds

  double RiceThreshold(double, double, double);

  unsigned int i, fs2=fs/2;
   
  //  eventsList->IncrementNowId();

  if(Nsum == 0) {

    if(mr) delete mr;
    if(mi) delete mi;
    if(s2r) delete s2r;
    if(s2i) delete s2i;

    mr = new double[fs2];
    mi = new double[fs2];
    s2r = new double[fs2];
    s2i = new double[fs2];

    memset(mr, 0, fs2*sizeof(double));
    memset(mi, 0, fs2*sizeof(double));
    memset(s2r, 0, fs2*sizeof(double));
    memset(s2i, 0, fs2*sizeof(double));
  }

  Nsum++;

  // get data
  Complex data[fs2];
  if(fs2 != s->getData(fs2,data))
    throw TidError ( "FSeries length doesn't match sampling frequency." );

  unsigned int nSteps;
  nSteps = (unsigned int)floor(stabScale / timeRes);

  for(i=1; i<fs2; i++) { // loop over frequency, skip DC

    mr[i] += data[i].Real();
    mi[i] += data[i].Imag();

    s2r[i] += data[i].Real() * data[i].Real();
    s2i[i] += data[i].Imag() * data[i].Imag();

  

    if(Nsum == nSteps) {

      mr[i] /= (double)nSteps;
      mi[i] /= (double)nSteps;
      s2r[i] = s2r[i] / (double)nSteps - mr[i]*mr[i];
      s2i[i] = s2i[i] / (double)nSteps - mi[i]*mi[i];

      MeanValues[i] = Complex(mr[i], mi[i]);
      VarValues[i] = Complex(s2r[i], s2i[i]);

      try {
	double Q = mr[i]*mr[i] + mi[i]*mi[i],
	  P = s2r[i] + s2i[i];

	if(Q==0.0 && P==0.0) {
	  throw TidError (name + string(": Null values -- can't compute threshold"));
	}
	
	if(Q/P > 1E6) 
	  throw TidError (name + string(": Line too sharp -- can't compute threshold"));

	powThresh[i] = RiceThreshold(bpp, Q, P);
      }
      catch (TidError err) {
        cerr<<"On frequency Id "<<i<<endl;
        HandleError(err);
	powThresh[i] = 1E50;
      }

      //      printf("%i:\t%g\n",i,powThresh[i]);
    }
  }

  if(Nsum == nSteps) {

    eventsList->Clear();

    eventsList->SetPThresh(powThresh);

    Nsum = 0;

    delete mr; mr = NULL;
    delete mi; mi = NULL;
    delete s2r; s2r = NULL;
    delete s2i;  s2i = NULL;

    mr = mi = s2r = s2i = NULL;
    thrCompMode = 0;
    ForceComputeThreshold = 0;
  }

}


double RiceThreshold(double bpp, double Q, double P0) {
  // computes threshold to apply on power to have black pixel probability
  // bpp, for a Rice distribution with scale P0 and non-centrality Q

  float rtsafe(void (*funcd)(float, float *, float *, double, double, double), 
	       float x1, float x2,
	       float xacc,
	       double P0, double Q, double bpp);
  void RiceWrapper(float P, float *f, float *df, double P0, double Q, double bpp);

  double P;

  P = rtsafe(RiceWrapper, P0+Q, P0+Q-2.0*log(bpp)*sqrt(P0*(P0+2.0*Q)), 1E-3, P0, Q, bpp);

  return P;

}

void RiceWrapper(float P, float *f, float *df, double P0, double Q, double bpp) {
  float Rice(float P, double Q, double P0);
  float RiceInt(float bpp, float P, double Q, double P0);

  *f = RiceInt(bpp, P, Q, P0);
  *df = Rice(P, Q, P0);
}

float Rice(float Pt, double Q, double P0) {
  // computes the value of the Rice distribution
  double i0(double x);
  double p, P = (double)Pt;

  p = i0(2.0*sqrt(P*Q)/P0) * exp(-(P+Q)/P0) / P0;

  return p;

}

float RiceInt(float bpp, float P, double Q, double P0) {
  // computes the error from goal bpp
  float qromb(float (*)(float, double, double),
	      float, float, double, double);
  float er;

  er = 1.0 - bpp - qromb(Rice, 0, P, Q, P0);

  return er;

}


bool ChnInfo::TestStationarity () { // test background noise for stationarity

  // H0: x ~ N(mx, s^2)
  //     y ~ N(my, s^2)
  // H1: mx, my, s^2 not equal to reference
  // returns 0 when H1 for any of the test frequencies

  float gammq(float, float);

  double chi2, buf;
  unsigned int i;

  unsigned int nTSteps = (unsigned int)floor(stabScale / timeRes);

  if(thrCompMode) return 1;

  if(History[0]->size() != nTSteps) return 1;

  bool FLAG = 0;

  for(i=0; i<testFreq.size(); i++) {

    // test compatibility with stored value using chi-2
    ListComplex::iterator p;
    for(p=History[i]->begin(); p!=History[i]->end(); p++) {
      buf = (*p).Real() - MeanValues[testFreq[i]].Real();
      chi2 += buf * buf / VarValues[testFreq[i]].Real();
      buf = (*p).Imag() - MeanValues[testFreq[i]].Imag();
      chi2 += buf * buf / VarValues[testFreq[i]].Imag();
    }
    chi2 /= 2.0 * (double)nTSteps;

    //    printf("%i\t\t%g\t\t%i\t\t%g\n",i,chi2,nTSteps,gammq(chi2, 2.0 * (double)nTSteps));
    if(gammq(chi2, 2.0 * (double)nTSteps) > stabTol)  // inconsistent result
      FLAG = 1;
  }

  // clear history buffer
  for(i=0; i<testFreq.size(); i++)
    History[i]->clear(); 

  if(FLAG) return 0;

  return 1;
}




/*****************NUMERICAL RECIPES CODE**************************/

float gammq(float a, float x)
{
	void gcf(float *gammcf, float a, float x, float *gln);
	void gser(float *gamser, float a, float x, float *gln);
	float gamser,gammcf,gln;

	if (x < 0.0 || a <= 0.0) 
	  throw TidError("Invalid arguments in routine gammq.");

	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return 1.0-gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return gammcf;
	}
}


#include <math.h>
#define ITMAX 1000
#define EPS 3.0e-7
#define FPMIN 1.0e-30

void gcf(float *gammcf, float a, float x, float *gln)
{
	float gammln(float xx);
	int i;
	float an,b,c,d,del,h;

	*gln=gammln(a);
	b=x+1.0-a;
	c=1.0/FPMIN;
	d=1.0/b;
	h=d;
	for (i=1;i<=ITMAX;i++) {
		an = -i*(i-a);
		b += 2.0;
		d=an*d+b;
		if (fabs(d) < FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c) < FPMIN) c=FPMIN;
		d=1.0/d;
		del=d*c;
		h *= del;
		if (fabs(del-1.0) < EPS) break;
	}
	if (i > ITMAX) 
	  throw TidError("a too large, ITMAX too small in gcf");
	*gammcf=exp(-x+a*log(x)-(*gln))*h;
}
#undef ITMAX
#undef EPS
#undef FPMIN



#include <math.h>
#define ITMAX 1000
#define EPS 3.0e-7

void gser(float *gamser, float a, float x, float *gln)
{
	float gammln(float xx);
	int n;
	float sum,del,ap;

	*gln=gammln(a);
	if (x <= 0.0) {
		if (x < 0.0) throw TidError("x less than 0 in routine gser");
		*gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=1;n<=ITMAX;n++) {
			++ap;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS) {
				*gamser=sum*exp(-x+a*log(x)-(*gln));
				return;
			}
		}
		throw TidError("a too large, ITMAX too small in routine gser");
		return;
	}
}
#undef ITMAX
#undef EPS



#include <math.h>

float gammln(float xx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}


/*****************************************************************/

#include <math.h>
#define EPS 1.0e-6
#define JMAX 30
#define JMAXP (JMAX+1)
#define K 5

float qromb(float (*func)(float, double, double), float a, float b,
	    double P0, double Q)
{
	void polint(float xa[], float ya[], int n, float x, float *y, float *dy);
	float trapzd(float (*func)(float, double, double), float a, float b, int n, double, double);

	float ss,dss;
	float s[JMAXP],h[JMAXP+1];
	int j;

	h[1]=1.0;
	for (j=1;j<=JMAX;j++) {
		s[j]=trapzd(func,a,b,j,P0,Q);
		if (j >= K) {
			polint(&h[j-K],&s[j-K],K,0.0,&ss,&dss);
			if (fabs(dss) <= EPS*fabs(ss)) return ss;
		}
		h[j+1]=0.25*h[j];
	}
	throw TidError("Too many steps in routine qromb.");
	return 0.0;
}
#undef EPS
#undef JMAX
#undef JMAXP
#undef K


#define FUNC(x) ((*func)(x, P0, Q))

float trapzd(float (*func)(float, double, double), float a, float b, int n,
	     double P0, double Q)
{
	float x,tnm,sum,del;
	static float s;
	int it,j;

	if (n == 1) {
		return (s=0.5*(b-a)*(FUNC(a)+FUNC(b)));
	} else {
		for (it=1,j=1;j<n-1;j++) it <<= 1;
		tnm=it;
		del=(b-a)/tnm;
		x=a+0.5*del;
		for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x);
		s=0.5*(s+(b-a)*sum/tnm);
		return s;
	}
}
#undef FUNC


#include <math.h>
#define NRANSI

void polint(float xa[], float ya[], int n, float x, float *y, float *dy)
{
  float *nrvector(long nl, long nh);
  void free_vector(float *v, long nl, long nh);

	int i,m,ns=1;
	float den,dif,dift,ho,hp,w;
	float *c,*d;

	dif=fabs(x-xa[1]);
	c=nrvector(1,n);
	d=nrvector(1,n);
	for (i=1;i<=n;i++) {
		if ( (dift=fabs(x-xa[i])) < dif) {
			ns=i;
			dif=dift;
		}
		c[i]=ya[i];
		d[i]=ya[i];
	}
	*y=ya[ns--];
	for (m=1;m<n;m++) {
		for (i=1;i<=n-m;i++) {
			ho=xa[i]-x;
			hp=xa[i+m]-x;
			w=c[i+1]-d[i];
			if ( (den=ho-hp) == 0.0) throw TidError("Error in routine polint");
			den=w/den;
			d[i]=hp*den;
			c[i]=ho*den;
		}
		*y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
	}
	free_vector(d,1,n);
	free_vector(c,1,n);
}
#undef NRANSI

#define NR_END 1
#define FREE_ARG char*

float *nrvector(long nl, long nh)
/* allocate a float vector with subscript range v[nl..nh] */
{
	float *v;

	v=(float *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(float)));
	if (!v) throw TidError("allocation failure in vector()");
	return v-nl+NR_END;
}

void free_vector(float *v, long nl, long nh)
/* free a float vector allocated with vector() */
{
	free((FREE_ARG) (v+nl-NR_END));
}

#undef NR_END
#undef FREE_ARG



/*********************************************************************/

#include <math.h>
#define MAXIT 1000

float rtsafe(void (*funcd)(float, float *, float *, double, double, double), 
	     float x1, float x2,
	     float xacc,
	     double P0, double Q, double bpp)
{
	int j;
	float df,dx,dxold,f,fh,fl;
	float temp,xh,xl,rts;

	(*funcd)(x1,&fl,&df,P0,Q,bpp);
	(*funcd)(x2,&fh,&df,P0,Q,bpp);
	if ((fl > 0.0 && fh > 0.0) || (fl < 0.0 && fh < 0.0))
		throw TidError("Root must be bracketed in rtsafe");
	if (fl == 0.0) return x1;
	if (fh == 0.0) return x2;
	if (fl < 0.0) {
		xl=x1;
		xh=x2;
	} else {
		xh=x1;
		xl=x2;
	}
	rts=0.5*(x1+x2);
	dxold=fabs(x2-x1);
	dx=dxold;
	(*funcd)(rts,&f,&df,P0,Q,bpp);
	for (j=1;j<=MAXIT;j++) {
		if ((((rts-xh)*df-f)*((rts-xl)*df-f) >= 0.0)
			|| (fabs(2.0*f) > fabs(dxold*df))) {
			dxold=dx;
			dx=0.5*(xh-xl);
			rts=xl+dx;
			if (xl == rts) return rts;
		} else {
			dxold=dx;
			dx=f/df;
			temp=rts;
			rts -= dx;
			if (temp == rts) return rts;
		}
		if (fabs(dx) < xacc) return rts;
		(*funcd)(rts,&f,&df,P0,Q,bpp);
		if (f < 0.0)
			xl=rts;
		else
			xh=rts;
	}
	throw TidError("Maximum number of iterations exceeded in rtsafe");
	return 0.0;
}
#undef MAXIT
